@@ -26,8 +26,12 @@ from . import (
     obsutil,
     util,
 )
+from .utils import (
+    stringutil,
+)
 
-def _bundle(repo, bases, heads, node, suffix, compress=True, obsolescence=True):
+def backupbundle(repo, bases, heads, node, suffix, compress=True,
+                 obsolescence=True):
     """create a bundle with the specified revisions as a backup"""
 
     backupdir = "strip-backup"
@@ -166,7 +170,7 @@ def strip(ui, repo, nodelist, backup=True, topic='backup'):
     vfs = repo.vfs
     node = nodelist[-1]
     if backup:
-        backupfile = _bundle(repo, stripbases, cl.heads(), node, topic)
+        backupfile = backupbundle(repo, stripbases, cl.heads(), node, topic)
         repo.ui.status(_("saved backup bundle to %s\n") %
                        vfs.join(backupfile))
         repo.ui.log("backupbundle", "saved backup bundle to %s\n",
@@ -179,78 +183,80 @@ def strip(ui, repo, nodelist, backup=True, topic='backup'):
         # we are trying to strip.  This is harmless since the stripped markers
         # are already backed up and we did not touched the markers for the
         # saved changesets.
-        tmpbundlefile = _bundle(repo, savebases, saveheads, node, 'temp',
-                                compress=False, obsolescence=False)
+        tmpbundlefile = backupbundle(repo, savebases, saveheads, node, 'temp',
+                                     compress=False, obsolescence=False)
 
-    try:
-        with repo.transaction("strip") as tr:
-            offset = len(tr.entries)
+    with ui.uninterruptable():
+        try:
+            with repo.transaction("strip") as tr:
+                offset = len(tr.entries)
 
-            tr.startgroup()
-            cl.strip(striprev, tr)
-            stripmanifest(repo, striprev, tr, files)
+                tr.startgroup()
+                cl.strip(striprev, tr)
+                stripmanifest(repo, striprev, tr, files)
 
-            for fn in files:
-                repo.file(fn).strip(striprev, tr)
-            tr.endgroup()
+                for fn in files:
+                    repo.file(fn).strip(striprev, tr)
+                tr.endgroup()
 
-            for i in xrange(offset, len(tr.entries)):
-                file, troffset, ignore = tr.entries[i]
-                with repo.svfs(file, 'a', checkambig=True) as fp:
-                    fp.truncate(troffset)
-                if troffset == 0:
-                    repo.store.markremoved(file)
+                for i in xrange(offset, len(tr.entries)):
+                    file, troffset, ignore = tr.entries[i]
+                    with repo.svfs(file, 'a', checkambig=True) as fp:
+                        fp.truncate(troffset)
+                    if troffset == 0:
+                        repo.store.markremoved(file)
 
-            deleteobsmarkers(repo.obsstore, stripobsidx)
-            del repo.obsstore
-            repo.invalidatevolatilesets()
-            repo._phasecache.filterunknown(repo)
+                deleteobsmarkers(repo.obsstore, stripobsidx)
+                del repo.obsstore
+                repo.invalidatevolatilesets()
+                repo._phasecache.filterunknown(repo)
 
-        if tmpbundlefile:
-            ui.note(_("adding branch\n"))
-            f = vfs.open(tmpbundlefile, "rb")
-            gen = exchange.readbundle(ui, f, tmpbundlefile, vfs)
-            if not repo.ui.verbose:
-                # silence internal shuffling chatter
-                repo.ui.pushbuffer()
-            tmpbundleurl = 'bundle:' + vfs.join(tmpbundlefile)
-            txnname = 'strip'
-            if not isinstance(gen, bundle2.unbundle20):
-                txnname = "strip\n%s" % util.hidepassword(tmpbundleurl)
-            with repo.transaction(txnname) as tr:
-                bundle2.applybundle(repo, gen, tr, source='strip',
-                                    url=tmpbundleurl)
-            if not repo.ui.verbose:
-                repo.ui.popbuffer()
-            f.close()
+            if tmpbundlefile:
+                ui.note(_("adding branch\n"))
+                f = vfs.open(tmpbundlefile, "rb")
+                gen = exchange.readbundle(ui, f, tmpbundlefile, vfs)
+                if not repo.ui.verbose:
+                    # silence internal shuffling chatter
+                    repo.ui.pushbuffer()
+                tmpbundleurl = 'bundle:' + vfs.join(tmpbundlefile)
+                txnname = 'strip'
+                if not isinstance(gen, bundle2.unbundle20):
+                    txnname = "strip\n%s" % util.hidepassword(tmpbundleurl)
+                with repo.transaction(txnname) as tr:
+                    bundle2.applybundle(repo, gen, tr, source='strip',
+                                        url=tmpbundleurl)
+                if not repo.ui.verbose:
+                    repo.ui.popbuffer()
+                f.close()
 
-        with repo.transaction('repair') as tr:
-            bmchanges = [(m, repo[newbmtarget].node()) for m in updatebm]
-            bm.applychanges(repo, tr, bmchanges)
+            with repo.transaction('repair') as tr:
+                bmchanges = [(m, repo[newbmtarget].node()) for m in updatebm]
+                bm.applychanges(repo, tr, bmchanges)
 
-        # remove undo files
-        for undovfs, undofile in repo.undofiles():
-            try:
-                undovfs.unlink(undofile)
-            except OSError as e:
-                if e.errno != errno.ENOENT:
-                    ui.warn(_('error removing %s: %s\n') %
-                            (undovfs.join(undofile), str(e)))
+            # remove undo files
+            for undovfs, undofile in repo.undofiles():
+                try:
+                    undovfs.unlink(undofile)
+                except OSError as e:
+                    if e.errno != errno.ENOENT:
+                        ui.warn(_('error removing %s: %s\n') %
+                                (undovfs.join(undofile),
+                                 stringutil.forcebytestr(e)))
 
-    except: # re-raises
-        if backupfile:
-            ui.warn(_("strip failed, backup bundle stored in '%s'\n")
-                    % vfs.join(backupfile))
-        if tmpbundlefile:
-            ui.warn(_("strip failed, unrecovered changes stored in '%s'\n")
-                    % vfs.join(tmpbundlefile))
-            ui.warn(_("(fix the problem, then recover the changesets with "
-                      "\"hg unbundle '%s'\")\n") % vfs.join(tmpbundlefile))
-        raise
-    else:
-        if tmpbundlefile:
-            # Remove temporary bundle only if there were no exceptions
-            vfs.unlink(tmpbundlefile)
+        except: # re-raises
+            if backupfile:
+                ui.warn(_("strip failed, backup bundle stored in '%s'\n")
+                        % vfs.join(backupfile))
+            if tmpbundlefile:
+                ui.warn(_("strip failed, unrecovered changes stored in '%s'\n")
+                        % vfs.join(tmpbundlefile))
+                ui.warn(_("(fix the problem, then recover the changesets with "
+                          "\"hg unbundle '%s'\")\n") % vfs.join(tmpbundlefile))
+            raise
+        else:
+            if tmpbundlefile:
+                # Remove temporary bundle only if there were no exceptions
+                vfs.unlink(tmpbundlefile)
 
     repo.destroyed()
     # return the backup file path (or None if 'backup' was False) so
@@ -350,10 +356,10 @@ def rebuildfncache(ui, repo):
         newentries = set()
         seenfiles = set()
 
-        repolen = len(repo)
+        progress = ui.makeprogress(_('rebuilding'), unit=_('changesets'),
+                                   total=len(repo))
         for rev in repo:
-            ui.progress(_('rebuilding'), rev, total=repolen,
-                        unit=_('changesets'))
+            progress.update(rev)
 
             ctx = repo[rev]
             for f in ctx.files():
@@ -370,7 +376,7 @@ def rebuildfncache(ui, repo):
                 if repo.store._exists(d):
                     newentries.add(d)
 
-        ui.progress(_('rebuilding'), None)
+        progress.complete()
 
         if 'treemanifest' in repo.requirements: # safe but unnecessary otherwise
             for dir in util.dirs(seenfiles):
@@ -400,18 +406,6 @@ def rebuildfncache(ui, repo):
         else:
             ui.write(_('fncache already up to date\n'))
 
-def stripbmrevset(repo, mark):
-    """
-    The revset to strip when strip is called with -B mark
-
-    Needs to live here so extensions can use it and wrap it even when strip is
-    not enabled or not present on a box.
-    """
-    return repo.revs("ancestors(bookmark(%s)) - "
-                     "ancestors(head() and not bookmark(%s)) - "
-                     "ancestors(bookmark() and not bookmark(%s))",
-                     mark, mark, mark)
-
 def deleteobsmarkers(obsstore, indices):
     """Delete some obsmarkers from obsstore and return how many were deleted
 
